#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "delete_lft_temporary_objects"

import sys
import os
import re
import getopt

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import Utility
from Utility import print_status, print_status_error
import HDIClient
import HDIOperation
import HCPClient
import HCPMQE

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions

# Regular expression
url_parser = re.compile(r"(http|https)://([^.]+)\.([^.]+)\.([^/]+)(/rest/.+)")

def usage():
    print
    print "Usage:"
    print "    sudo maintexec %s --namespace <namespace name>"

def parse_url(url):
    match = url_parser.match(url)
    if match is None:
        return (None, None, None, None)
    ns = match.group(2)
    t = match.group(3)
    d = match.group(4)
    obj = match.group(5)

    return (ns, t, d, obj)

def delete_lft_temporary_objects(namespace, hcp_info, config):
    ### Find temporary files (UUID.tmp) with MQE
    log_info("--- Look up temporary objects on %s ---" % namespace)

    # Setup MQE query
    hcp_client = HCPClient.HCPClient(hcp_info)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    hcp_client.enable_content_header()
    query = HCPMQE.HCPMQEObjectQuery()
    query.set_HCPClient(hcp_client)
    queryexp = "namespace:%s.%s AND objectPath:/*/*/*.tmp" % (namespace, hcp_info.tenant)
    query.set_query(queryexp)
    print_status("Getting total count of temporary objects...")

    # Count total objects at first
    total_count = query.get_total_count()
    log_info("Total %s temporary objects." % total_count)
    if total_count == 0:
        print_status("No temporary objects found.")
        return

    print_status("%s temporary objects found." % total_count)

    # Get result every 100 objects, and store all of the objects information on memory
    # assuming there are not many temporary objects...
    total_count = 0 # Reuse
    urllist = []
    query.set_count(100)
    print_status("Deleting the temporary objects ...")
    while not query.completed:
        queryXML = query.build()
        log_debug("MQE Query: %s" % queryXML)

        # Issue MQE query
        log_info("Issuing query...")
        try:
            resultset = query.issue(build=False)
        except HCPMQE.ConnectionError, e:
            raise Exceptions.ConnectionError(e.reason, e.code)
        except HCPMQE.MQEError, e:
            log_error("An unexpected error occurred while executing MQE query. error=%s" % str(e))
            raise Exceptions.OperationError("MQE failure. error=%s" % e.reason, e.code)

        # Parse the result
        log_info("Got query result.")
        objects = resultset.findall("object")
        if len(objects) == 0:
            # This shouldn't be 0 because the status was "INCOMPLETE"
            log_warning("0 objects were returned?")
            log_warning("MQE query=%s" % queryXML)
        total_count += len(objects)
        log_info("%s results were returned. Total %s results" % (len(objects), total_count))

        # Remember path of the temporary objects
        for obj in objects:
            url = obj.attrib['urlName']
            # Sanity check
            if not url.endswith(".tmp"):
                log_warning("The result doesn't look like a temporary object? url=%s" % url)
                continue
            urllist.append(url)

    ### Delete the found temporary objects
    print_status("Deleting the found temporary objects ...")
    hcp_client = HCPClient.HCPClient(hcp_info, namespace)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)

    for idx, url in enumerate(urllist):
        log_info("Deleting %s... (remaining %s)" % (url, len(urllist) - idx))
        ns, tn, dm, obj_path = parse_url(url)

        # Sanity check again
        if not obj_path.endswith(".tmp"):
            log_warning("The object path does not end with .tmp. Skip this object. path=%s" % obj_path)
            continue

        # Delete
        try:
            ret = hcp_client.delete(obj_path)
        except HCPClient.ConnectionError, e:
            log_error("Failed to connect HCP. error=%s" % str(e))
            raise Exceptions.ConnectionError(e.reason, e.code)

        if ret['code'] == 404:
            log_info("404 was returned when deleting an object. obj=%s" % url)
            continue
        elif ret['code'] == 204:
            log_info("204 was returned when deleting an object. obj=%s" % url)
            continue
        elif ret['code'] != 200:
            if 'X-HCP-ErrorMessage' in ret['headers']:
                errmsg = ret['headers']['X-HCP-ErrorMessage']
            else:
                errmsg = "None"
            log_error("Unexpected HTTP status code. code=%s, msg=%s, error=%s" % (ret['code'], ret['message'], errmsg))
            log_error("Failed to delete a temporary object on the namespace. obj=%s, namespace=%s" % (obj_path, namespace))
            print_status_error("Could not delete a temporary object. Continue to delete other objects ...")

    print_status("Deletion completed.")

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Parse options
    short_opts = ''
    long_opts = ['namespace=']
    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        Exceptions.OptionError(e)

    namespace = None
    for o, a in opts:
        if o == '--namespace':
            namespace = a

    if namespace is None:
        raise Exceptions.OptionError("--namespace option is required.")

    # Prepare connection to localhost
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    # Get HCP information
    print_status("Loading HCP configuration ...")
    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. error=%s" % str(e))
        raise Exceptions.OperationError("Failed to get HCP information on the node. error=%s" % e.reason, e.code)

    # Delete temporary objects of LFT on the namespace
    delete_lft_temporary_objects(namespace, hcp_info, config)

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option Error: %s" % str(e))
        usage()
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
