#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "failover_hcp"

import sys
import os
import time
import csv
import errno
import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from Utility import print_status, print_status_warning, print_status_error
import HCPClient
import HDIClient
import HDIOperation
import FailoverConfig

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger
import setup_common

# Constants

last_info_file = "replication.csv"
saved_replica_user = None
saved_replica_pass = None

def load_last_information(config):
    last_info_file_path = os.path.join(config.fo_conf_dir, last_info_file)
    lastinfo = {}
    try:
        with open(last_info_file_path, 'r') as f:
            reader = csv.reader(f)
            items = reader.next()
            lastinfo['domain'] = items[0]
            lastinfo['link'] = items[1]
            lastinfo['exhost'] = items[2] if len(items[2]) != 0 else None
    except StopIteration:
        print_status_warning("No lines are recorded in the record of HCP replication information. path=%s" % last_info_file_path)
        lastinfo = None
    except EnvironmentError, e:
        if e.errno == errno.ENOENT:
            log_info("HCP replication information at the last execution does not exist.")
        else:
            print_status_warning("Failed to read HCP replication information at the last execution. error=%s" % str(e))
        lastinfo = None

    return lastinfo

def save_last_information(link, hcp_info, config):
    last_info_file_path = os.path.join(config.fo_conf_dir, last_info_file)
    try:
        with open(last_info_file_path, 'w') as f:
            writer = csv.writer(f, lineterminator='\n')
            writer.writerow([hcp_info.domain, link, hcp_info.exhost])
    except EnvironmentError, e:
        print_status_error("Failed to save HCP replica information. error=%s" % str(e))
        raise

def get_system_admin_account(reset=False):
    global saved_replica_user
    global saved_replica_pass

    if not reset and saved_replica_user is not None:
        return (saved_replica_user, saved_replica_pass)

    print
    print ">>> Enter HCP system-level administrator account on the replica HCP."
    print
    saved_replica_user = Utility.my_input("Name: ")
    if saved_replica_user is None:
        raise Exceptions.NoExecution()
    saved_replica_pass = Utility.my_input_secret("Password: ")
    if saved_replica_pass is None:
        raise Exceptions.NoExecution()

    return (saved_replica_user, saved_replica_pass)

def get_replica_hcp_information(config):
    """ Get HCP replica information.

    This function returns:
        HCPInfo for operating HCP replication
        HCP replication link name to operate

    If there is the information at the last execution, this function loads it
    and returns it.
    If not or if a user choose to enter a new information, this function gets
    the information from the user and returns it.
    """

    # Get HCP information
    print_status("Getting HCP information defined on this node ...")
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. host=%s, error=%s" % (conn.host, str(e)))
        raise Exceptions.OperationError("Failed to get HCP settings. error=%s" % e.reason, e.code)

    # Load the last information if exists
    lastinfo = load_last_information(config)
    if lastinfo is not None:
        print_status("Information used at the last execution is found.")
        log_info("lastinfo: %s" % lastinfo)
        display_hcp_failover_configuration(lastinfo['domain'], lastinfo['link'])
        # Ask if the admin wants to use this information (or re-enter another information)
        answer = Utility.yes_or_no("Do you want to re-use this information?")
        if answer is True:
            replica_user, replica_pass = get_system_admin_account()
            hcp_info_to_failover = setup_hcp_info_to_failover(lastinfo['domain'], replica_user, replica_pass, lastinfo['exhost'], hcp_info)
            return (hcp_info_to_failover, lastinfo['link'])

    # Get a new information

    primary = hcp_info.domain
    replica = hcp_info.rephost

    # Select Replica HCP
    PRIMARY = 1
    REPLICA = 2
    hcp_menu = []
    hcp_menu.append((PRIMARY, primary))
    if replica is not None:
        hcp_menu.append((REPLICA, replica))

    print
    print ">>> Enter information of *replica* HCP to operate."
    selected_list = Utility.select_items("Select a *REPLICA* HCP that you want to operate", hcp_menu, title="Defined HCP domains", onlyone=True)
    if selected_list is None:
        log_info("Nothing is selected as replica HCP.")
        raise Exceptions.NoExecution()

    selected = selected_list[0]
    if selected == PRIMARY:
        hcp_to_failover = primary
        exhost = hcp_info.exhost
    elif selected == REPLICA:
        hcp_to_failover = replica
        exhost = hcp_info.repexhost
    log_info("Selected HCP: %s" % hcp_to_failover)

    # Get system admin account of the replica HCP
    replica_user, replica_pass = get_system_admin_account(reset=True)

    # Setup HCPInfo to issue MAPI for the HCP system
    hcp_info_to_failover = setup_hcp_info_to_failover(hcp_to_failover, replica_user, replica_pass, exhost, hcp_info)
    # Get replication links defined on the replica HCP
    links = list_replication_links(hcp_info_to_failover, config)

    # Select a replication link to execute failover
    link_menu = []
    for i, link in enumerate(links):
        idx = i + 1
        link_menu.append((idx, link))

    print
    print ">>> Select a replication link that you want to execute failover."
    selected_list = Utility.select_items("Select a replication link to execute failover", link_menu, title="Defined replication links on the replica HCP", onlyone=True)
    if selected_list is None:
        log_info("Nothing is selected as replication link.")
        raise Exceptions.NoExecution()

    selected = selected_list[0]
    selected_link = links[idx - 1]

    # Save this information for use at the next execution
    save_last_information(selected_link, hcp_info_to_failover, config)

    return (hcp_info_to_failover, selected_link)

def display_hcp_failover_configuration(hcp_to_failover, link):
    print
    print "----- HCP failover configuration -----"
    print
    print "Replica HCP:"
    print "\t%s" % hcp_to_failover
    print "Replication link:"
    print "\t%s" % link
    print
    print "--------------------------------------"
    print

def setup_hcp_info_to_failover(hcp_to_failover, replica_user, replica_pass, exhost, hcp_info):
    hcp_info_to_failover = HCPClient.HCPInfo()
    hcp_info_to_failover.set_domain(hcp_to_failover)
    hcp_info_to_failover.set_tenant("admin")
    hcp_info_to_failover.set_account(replica_user, password=replica_pass)
    hcp_info_to_failover.set_exhost(exhost)
    if hcp_info.proxy is not None:
        hcp_info_to_failover.set_proxy(proxy=hcp_info.proxy)

    return hcp_info_to_failover

def execute_hcp_failover(link, hcp_info, config):
    # Setup HCPClient
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)

    # Issue MAPI to get list of replication links
    path = "/mapi/services/replication/links/%s" % link
    params = {'failOver': None}
    print_status("Executing failover on %s ..." % link)
    try:
        hcp_client.set_params(params)
        ret = hcp_client.post(path)
    except HCPClient.ConnectionError, e:
        print_status_error("Failed to connect HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        print_status_error("Failed to execute failover on replication link %s on HCP. error=%s" % (link, ret['message']))
        raise Exceptions.OperationError("Failed to execute HCP failover.", ret['code'])

    print_status("Failover has been triggered.")

def execute_begin_data_recovery(link, hcp_info, config):
    # Setup HCPClient
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)

    # Issue MAPI to get list of replication links
    path = "/mapi/services/replication/links/%s" % link
    params = {'beginRecovery': None}
    print_status("Executing 'begin data recovery' on %s ..." % link)
    try:
        hcp_client.set_params(params)
        ret = hcp_client.post(path)
    except HCPClient.ConnectionError, e:
        print_status_error("Failed to connect HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        print_status_error("Failed to execute 'begin data recovery' on replication link %s on HCP. error=%s" % (link, ret['message']))
        raise Exceptions.OperationError("Failed to execute 'begin data recovery' on HCP.", ret['code'])

    print_status("Data recovery has started.")

def execute_complete_data_recovery(link, hcp_info, config):
    # Setup HCPClient
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)

    # Issue MAPI to get list of replication links
    path = "/mapi/services/replication/links/%s" % link
    params = {'completeRecovery': None}
    print_status("Executing 'complete data recovery' on %s ..." % link)
    try:
        hcp_client.set_params(params)
        ret = hcp_client.post(path)
    except HCPClient.ConnectionError, e:
        print_status_error("Failed to connect HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        print_status_error("Failed to execute 'complete data recovery' on replication link %s on HCP. error=%s" % (link, ret['message']))
        raise Exceptions.OperationError("Failed to execute 'complete data recovery' on HCP.", ret['code'])

    print_status("Completion of data recovery has started.")

def failover_hcp(config):
    # Get replica HCP information to operate
    hcp_info_to_failover, link = get_replica_hcp_information(config)

    # Confirmation
    display_hcp_failover_configuration(hcp_info_to_failover.domain, link)
    answer = Utility.yes_or_no_full("Are you ready to start HCP failover?")
    if answer is False:
        raise Exceptions.NoExecution()

    # Do nothing if it's already failed over
    print_status("Getting current replication status ...")
    status = get_replication_status(link, hcp_info_to_failover, config)
    display_link_status(status)
    if status['statusMessage'] == 'Failed over':
        print_status("The replication link status is already 'Failed over'.")
        return

    # Start HCP failover
    execute_hcp_failover(link, hcp_info_to_failover, config)

    # Check status
    print_status("Waiting until the link status becomes 'Failed over' ...")
    for i in range(0, 10):
        status = get_replication_status(link, hcp_info_to_failover, config)
        display_link_status(status)
        if status['statusMessage'] == 'Failed over':
            print_status("HCP failover has completed.")
            break
        time.sleep(10)
    else:
        print_status_warning("The link status did not become 'Failed over' after 100 seconds.")


def begin_data_recovery(config):
    # Get replica HCP information to operate
    hcp_info_to_failover, link = get_replica_hcp_information(config)

    # Confirmation
    display_hcp_failover_configuration(hcp_info_to_failover.domain, link)
    answer = Utility.yes_or_no_full("Are you ready to begin data recovery?")
    if answer is False:
        raise Exceptions.NoExecution()

    # Start data recovery
    execute_begin_data_recovery(link, hcp_info_to_failover, config)

    # Do not check the status here.
    # Expect a user to check the status by themselves.

def complete_data_recovery(config):
    # Get replica HCP information to operate
    hcp_info_to_failover, link = get_replica_hcp_information(config)

    # Confirmation
    display_hcp_failover_configuration(hcp_info_to_failover.domain, link)
    answer = Utility.yes_or_no_full("Are you ready to complete data recovery?")
    if answer is False:
        raise Exceptions.NoExecution()

    # Complete data recovery
    execute_complete_data_recovery(link, hcp_info_to_failover, config)

    # Do not check the status here.
    # Expect a user to check the status by themselves.

def monitor_link_status(config):
    # Get replica HCP information to operate
    hcp_info_to_failover, link = get_replica_hcp_information(config)

    # Check status
    INTERVAL = 15
    try:
        while True:
            status = get_replication_status(link, hcp_info_to_failover, config)
            print_status("Replication link status:")
            display_link_status_all(status)
            print
            print "Will update status %s seconds later ..." % INTERVAL
            print "Press Ctrl-C to stop monitoring."
            time.sleep(INTERVAL)
    except KeyboardInterrupt:
        # Just finish
        print_status("Interrupted.")

def swap_hcp_domain_on_hdi(primary, nodes, config, ssh_config):
    """ Swap primary HCP and replica HCP on HDI nodes.

    This function swaps primary HCP and replica HCP defined on each HDI node
    specified by 'nodes'. If primary HCP defined on a node is already primary
    HCP specified by 'primary', this function does not swap the HCPs.
    If none of primary nor replica HCP on a node matches 'primary', this function
    does not do anything on the node.

    Args:
        primary (str): HCP domain name that you want to define as primary HCP
            on each HDI node.
        nodes (list of str): List of IP address of HDI nodes that you want to
            swap HCP definition.
        config (Config.Config): Configuration of the scripts.
        ssh_config (Config.SSHConfig): SSH configuration.

    Returns:
        None
    """
    for node in nodes:
        cred = HDIClient.Credential(ssh_config.ssh_user, None, ssh_config.get_private_key(ipaddr=node), None)
        conn = HDIOperation.Connection(node, cred, config.retry_count, config.retry_interval, config.timeout)

        prog = os.path.join(BASEDIR, 'int/swap_hcp_domain')
        cmd = "sudo maintexec %s --primary %s" % (prog, primary)
        op = HDIOperation.CommandOperation(cmd)
        print_status("Setting %s as primary HCP on %s ..." % (primary, node))
        try:
            ret, out, err = op.execute(conn)
        except HDIOperation.ConnectionError, e:
            print_status_error("Failed to connect HDI %s. error=%s" % (conn.host, str(e)))
            print_status_error("Please execute %s after you removed the cause of the connection failure, to make sure all HDIs are in sync.")
            # Continue other HDIs

        if ret != 0:
            print_status_error("Failed to set the specified HCP domain on %s. ret=%s, out=%s, err=%s" % (node, ret, out, err))
            print_status_error("Please execute %s after you removed the cause of the connection failure, to make sure all HDIs are in sync." % PROGRAM)
            # Continue other HDIs

def change_hcp_on_hdi(config, nodes=None):
    # Get HCP information
    print_status("Getting HCP information defined on this node ...")
    cred = HDIClient.Credential(None, config.rest_user, None, config.rest_passMD5)
    conn = HDIOperation.Connection('localhost', cred, config.retry_count, config.retry_interval, config.timeout)

    try:
        hcpop = HDIOperation.HCPSetting(conn)
        hcp_info = hcpop.get()
    except HDIOperation.OperationError, e:
        log_error("Failed to get HCP information on the node. host=%s, error=%s" % (conn.host, str(e)))
        raise Exceptions.OperationError("Failed to get HCP settings. error=%s" % e.reason, e.code)

    primary = hcp_info.domain
    replica = hcp_info.rephost

    # Select HCP domain to be set as primary
    print
    print "HCP setting on this node:"
    print "\tPrimary: %s" % primary
    print "\tReplica: %s" % replica if replica is not None else "(Not set)"

    PRIMARY = 1
    REPLICA = 2
    EXIT = 0
    menu = [(PRIMARY, primary)]
    if replica is not None:
        menu.append((REPLICA, replica))
    menu.append((EXIT, "Cancel"))
    print
    print ">>> Select an HCP domain that you want to set as 'primary' on each HDI nodes."
    selected_list = Utility.select_items("Select an HCP domain", menu, title="Defined HCP domains", onlyone=True)
    if selected_list is None:
        log_info("Nothing is selected for HCP domain.")
        raise Exceptions.NoExecution()

    selected = selected_list[0]
    if selected == PRIMARY:
        hcp_to_be_primary = primary
    elif selected == REPLICA:
        hcp_to_be_primary = replica
    elif selected == EXIT:
        raise Exceptions.NoExecution()

    # Load the existing failover configuration
    log_info("Getting list of failover configuration.")
    try:
        confgroup_list = FailoverConfig.list_configuration_group(config.fo_conf_dir)
    except FailoverConfig.ConfigurationError, e:
        log_error("Failed to count number of configurations due to an configuration error. error=%s" % str(e))
        raise Exceptions.ConfigurationError(str(e))

    # Select a configuration group
    confgroup = setup_common.select_configuration_group(confgroup_list)

    # Load pair configuration and pair group configuration
    pairconf = FailoverConfig.PairConfig(config.fo_conf_dir, confgroup)

    # Setup SSHConfig for the configuration group
    ssh_config = Config.SSHConfig(config.sys_conf_dir, pairconf, confgroup)

    if nodes is None:
        # Select a pair(s) to change HCP setting
        print
        print ">>> Select a pair(s) to change HCP setting."
        pid_list = setup_common.select_pair(pairconf, msg="Select a pair(s) (separate with comma for multiple selection)", onlyone=False)
        if pid_list is None:
            print
            print "No pairs are selected."
            log_info("No pairs are selected.")
            raise Exceptions.NoExecution()

        # Select nodes to change HCP setting
        PRIMARY = 1
        SECONDARY = 2
        ALL = 0
        menu = [
            (PRIMARY, "Primary nodes in the selected pairs"),
            (SECONDARY, "Secondary nodes in the selected pairs"),
            (ALL, "All nodes in the selected pairs")
        ]
        print
        print ">>> Select nodes to change HCP setting."
        selected_list = Utility.select_items("Select nodes to change HCP setting", menu, onlyone=True)
        if selected_list is None:
            log_info("Nothing is selected for nodes to change HCP setting.")
            raise Exceptions.NoExecution()
        selected = selected_list[0]
        nodes = []
        for pid in pid_list:
            pair_list = pairconf.get_list(pair_id=pid)
            notused, pair = pair_list[0]
            if selected == PRIMARY or selected == ALL:
                nodes.append(pair.primary)
            if selected == SECONDARY or selected == ALL:
                nodes.append(pair.secondary)

    # Confirmation
    log_info("Nodes selected for updating HCP setting: %s" % nodes)
    print
    print "Selected nodes:"
    for node in nodes:
        print "\t%s" % node
    print
    print "HCP domain to be set as 'primary' on each HDI node:"
    print "\t%s" % hcp_to_be_primary
    answer = Utility.yes_or_no_full("Are you ready to change HCP setting on HDIs?")
    if answer is False:
        raise Exceptions.NoExecution()

    swap_hcp_domain_on_hdi(hcp_to_be_primary, nodes, config, ssh_config)

def get_replication_status(link, hcp_info, config):
    # Setup HCPClient
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    hcp_client.enable_content_header()

    # Issue MAPI to get list of replication links
    print_status("Getting replication link status of %s on the HCP ..." % link)
    path = "/mapi/services/replication/links/%s" % link
    try:
        # Need to enable verbose to get detail information
        hcp_client.enable_verbose()
        ret = hcp_client.get(path)
    except HCPClient.ConnectionError, e:
        print_status_error("Failed to connect HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        print_status_error("Failed to get list of replication link on HCP.")
        raise Exceptions.OperationError("Failed to get list of replication link on HCP.", ret['code'])

    # Parse the returned XML
    status = parse_link_status(ret['body'])

    return status

def display_link_status(status):
    # Display only status
    print
    print "----------------------------"
    print "status: %s" % status['status']
    print "statusMessage: %s" % status['statusMessage']
    print "suspended: %s" % status['suspended']
    print "----------------------------"

def display_link_status_all(status):
    print
    print "----------------------------"
    for key in sorted(status.keys()):
        if key == 'statistics':
            print "statistics:"
            for key2 in sorted(status['statistics'].keys()):
                print "\t%s: %s" % (key2, status['statistics'][key2])
        else:
            print "%s: %s" % (key, status[key])
    print "----------------------------"

def parse_links(xml_links):
    try:
        root = ET.fromstring(xml_links)
    except ExpatError, e:
        log_error("Invalid XML syntax in the provided XML. code=%s, line=%s, offset=%s, xml_data=%s" % (e.code, e.lineno, e.offset, xml_links))
        print_status_error("Failed to parse list of replication links returned from HCP.")
        raise Exceptions.OperationError("Unexpected XML contents returned from HCP.", -1)

    # root element must be '<links>'
    names = root.findall('name')
    return [name.text for name in names]

def parse_link_status(xml_link):
    # Get XML element from string
    try:
        root = ET.fromstring(xml_link)
    except ExpatError, e:
        log_error("Invalid XML syntax in the provided XML. code=%s, line=%s, offset=%s, xml_data=%s" % (e.code, e.lineno, e.offset, xml_link))
        print_status_error("Failed to parse replication link status returned from HCP.")
        raise Exceptions.OperationError("Unexpected XML contents returned from HCP.", -1)

    status = {}
    try:
        # root element must be '<link>'
        status['id'] = root.find('id').text
        status['name'] = root.find('name').text
        status['statistics'] = {}
        statistics = root.find('statistics')
        status['statistics']['upToDateAsOfString'] = statistics.find('upToDateAsOfString').text
        status['statistics']['upToDateAsOfMillis'] = float(statistics.find('upToDateAsOfMillis').text)
        status['statistics']['bytesPending'] = int(statistics.find('bytesPending').text)
        status['statistics']['bytesPendingRemote'] = int(statistics.find('bytesPendingRemote').text)
        status['statistics']['bytesPerSecond'] = float(statistics.find('bytesPerSecond').text)
        status['statistics']['bytesReplicated'] = int(statistics.find('bytesReplicated').text)
        status['statistics']['objectsPending'] = int(statistics.find('objectsPending').text)
        status['statistics']['objectsPendingRemote'] = int(statistics.find('objectsPendingRemote').text)
        status['statistics']['objectsReplicated'] = int(statistics.find('objectsReplicated').text)
        status['statistics']['operationsPerSecond'] = float(statistics.find('operationsPerSecond').text)
        status['statistics']['errors'] = int(statistics.find('errors').text)
        status['statistics']['errorsPerSecond'] = float(statistics.find('errorsPerSecond').text)
        status['status'] = root.find('status').text
        status['statusMessage'] = root.find('statusMessage').text
        status['suspended'] = Utility.str2bool(root.find('suspended').text)
        status['type'] = root.find('type').text
    except (AttributeError, ValueError), e:
        print_status_error("Unexpected values in replication link status. error=%s" % str(e))
        raise Exceptions.OperationError("Unexpected values in replication link status.", -1)

    return status

def list_replication_links(hcp_info, config):
    # Setup HCPClient
    hcp_client = HCPClient.HCPClient(hcp_info, port=9090)
    hcp_client.set_connection(config.retry_count, config.retry_interval, config.timeout)
    hcp_client.enable_content_header()

    # Issue MAPI to get list of replication links
    print_status("Getting replication links on the HCP ...")
    path = "/mapi/services/replication/links"
    try:
        ret = hcp_client.get(path)
    except HCPClient.ConnectionError, e:
        print_status_error("Failed to connect HCP. error=%s" % str(e))
        raise Exceptions.ConnectionError(e.reason, e.code)

    if not Utility.is_expected_HTTP_status(ret, (200,)):
        print_status_error("Failed to get list of replication link on HCP.")
        raise Exceptions.OperationError("Failed to get list of replication link on HCP.", ret['code'])

    # Parse XML contents and get list of replication link names
    links = parse_links(ret['body'])

    return links

def main():
    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configs
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Main menu
    OP_FAILOVER = 1
    OP_BEGIN_RECOVERY = 2
    OP_COMPLETE_RECOVERY = 3
    OP_MONITOR_STATUS = 4
    OP_CHANGE_HCP = 5
    OP_EXIT = 0

    menu = [
        (OP_FAILOVER, "Failover to replica HCP"),
        (OP_BEGIN_RECOVERY, "Begin data recovery to primary HCP"),
        (OP_COMPLETE_RECOVERY, "Complete data recovery to primary HCP"),
        (OP_MONITOR_STATUS, "Monitor replication link status"),
        (OP_CHANGE_HCP, "Change primary HCP setting on HDI"),
        (OP_EXIT, "Exit")
    ]

    while True:
        print
        print ">>> Select an operation."
        selected_list = Utility.select_items("Select an operation", menu, title="Operations", onlyone=True)
        if selected_list is None:
            raise Exceptions.NoExecution()
        selected = selected_list[0]
        try:
            if selected == OP_FAILOVER:
                failover_hcp(config)
            elif selected == OP_BEGIN_RECOVERY:
                begin_data_recovery(config)
            elif selected == OP_COMPLETE_RECOVERY:
                complete_data_recovery(config)
            elif selected == OP_MONITOR_STATUS:
                monitor_link_status(config)
            elif selected == OP_CHANGE_HCP:
                change_hcp_on_hdi(config)
            elif selected == OP_EXIT:
                log_info("Exit selected.")
                break
        except Exceptions.NoExecution:
            print "Going back to the main menu ..."

if __name__ == '__main__':
    try:
        main()
    except Exceptions.NoExecution:
        print "Process canceled."
        sys.exit(0)
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        print "Interrupted."
        log_warning("Interrupted.")
        sys.exit(2)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.ConfigurationError, e:
        log_error("Unexpected configuration: %s" % str(e))
        sys.exit(20)
    except Exceptions.ConnectionError, e:
        log_error("A connection error occurred: %s" % str(e))
        sys.exit(70)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(90)
    except EnvironmentError, e:
        log_exception("A system error occurred: %s", e)
        sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
