#!/usr/bin/python
# All Rights Reserved. Copyright (c) 2018, Hitachi, Ltd.       #
# Licensed Material of Hitachi, Ltd.                           #
# Reproduction, use, modification or disclosure otherwise than #
# permitted in the License Agreement is strictly prohibited.   #

from __future__ import with_statement
import xml.etree.ElementTree as ET
from xml.parsers.expat import ExpatError
from CommonLogger import log_debug, log_info, log_warning, log_error
import HCPClient

class BaseError(Exception):
    def __init__(self, reason, code=None):
        self.reason = reason.strip()
        self.code = code

    def __str__(self):
        if self.code is None:
            return "%s" % self.reason
        else:
            return "%s (code=%s)" % (self.reason, self.code)

class ParameterError(BaseError):
    pass

class MQEError(BaseError):
    pass

class ConnectionError(BaseError):
    pass

class HCPMQEOperationQuery(object):
    class LastResultEntry(object):
        def __init__(self, url, ctime, version):
            self.url = url
            self.ctime = ctime
            self.version = version

    def __init__(self):
        self.count = -1
        self.start_msec = None
        self.end_msec = None
        self.namespaces = []
        self.transactions = []
        self.dirs = []
        self.object_properties = []
        self.verbose = False
        self.indexiable = None
        self.completed = False
        self.last_result = None
        self.hcpclient = None
        self.query = None
        self.result = None
        self.status = None

    def set_HCPClient(self, hcpclient):
        self.hcpclient = hcpclient

    def set_count(self, count):
        self.count = count

    def set_starttime(self, start_msec):
        self.start_msec = start_msec

    def set_endtime(self, end_msec):
        self.end_msec = end_msec

    def enable_indexiable(self, value):
        # value must be a boolean value (True or False)
        self.indexiable = value

    def disable_indexiable(self):
        self.indexiable = None

    def enable_verbose(self):
        self.verbose = True

    def disable_verbose(self):
        self.verbose = False

    def add_namespace(self, ns):
        self.namespaces.append(ns)

    def clear_namespace(self):
        self.namespaces = []

    def add_transaction(self, t):
        self.transactions.append(t)

    def clear_transaction(self):
        self.transactions = []

    def add_directory(self, d):
        self.dirs.append(d)

    def clear_directory(self):
        self.dirs = []

    def add_property(self, p):
        self.object_properties.append(p)

    def clear_propery(self):
        self.object_properties = []

    def set_last_result(self, url, ctime, version):
        self.last_result = HCPMQEOperationQuery.LastResultEntry(url, ctime, version)

    def get_last_result(self):
        if self.last_result is None:
            return (None, None, None)
        else:
            return (self.last_result.url, self.last_result.ctime, self.last_result.version)

    def build(self):
        # Top-level entry
        top = ET.Element("queryRequest")

        # operation entry
        op = ET.SubElement(top, "operation")

        count = ET.SubElement(op, "count")
        count.text = str(self.count)
        verbose = ET.SubElement(op, "verbose")
        verbose.text = str(self.verbose).lower()

        # lastResult entry
        if self.completed is False and not self.last_result is None:
            last = ET.SubElement(op, "lastResult")
            lasturl = ET.SubElement(last, "urlName")
            lasturl.text = self.last_result.url
            lastctime = ET.SubElement(last, "changeTimeMilliseconds")
            lastctime.text = self.last_result.ctime
            lastversion = ET.SubElement(last, "version")
            lastversion.text = str(self.last_result.version)

        # objectProperties
        if len(self.object_properties) != 0:
            obj_prop = ET.SubElement(op, "objectProperties")
            properties = ','.join(self.object_properties)
            obj_prop.text = properties

        # systemMetadata entry
        sysmd = ET.SubElement(op, "systemMetadata")

        ctime = ET.SubElement(sysmd, "changeTime")
        if not self.start_msec is None:
            # start will be 0 if omit
            start = ET.SubElement(ctime, "start")
            start.text = str(self.start_msec)

        if not self.end_msec is None:
            # end will be "one minute before the time HCP received the request" if omit
            end = ET.SubElement(ctime, "end")
            end.text = str(self.end_msec)

        if len(self.dirs) != 0:
            dirs = ET.SubElement(sysmd, "directories")
            for dr in self.dirs:
                d = ET.SubElement(dirs, "directory")
                d.text = dr

        if not self.indexiable is None:
            idx = ET.SubElement(sysmd, "indexable")
            idx.text = str(self.indexiable).lower()

        if len(self.namespaces) == 0:
            # Force to specify at least one namespace
            raise ParameterError("At least one namespace has to be specified.")
        if self.hcpclient is None:
            # hcpclient needs to be set to get the tenant name
            raise ParameterError("HCP Connection has to be configured by set_HCPConnection()")
        nss = ET.SubElement(sysmd, "namespaces")
        for namespace in self.namespaces:
            ns = ET.SubElement(nss, "namespace")
            ns.text = namespace + "." + self.hcpclient.tenant

        if len(self.transactions) != 0:
            trans = ET.SubElement(sysmd, "transactions")
            for transaction in self.transactions:
                t = ET.SubElement(trans, "transaction")
                t.text = transaction

        self.query = ET.tostring(top)
        return self.query

    def issue(self, build=True):
        if self.query is None or build is True:
            self.build()
        try:
            ret = self.hcpclient.exec_MQE(self.query)
        except HCPClient.ConnectionError, e:
            log_error("Connection failed on executing MQE. error=%s" % str(e))
            raise ConnectionError(e.reason, e.code)

        if ret['code'] != 200:
            if 'X-HCP-ErrorMessage' in ret['headers']:
                errmsg = ret['headers']['X-HCP-ErrorMessage']
            else:
                errmsg = "None"
            raise MQEError("MQE failure: HTTP return code=%s, msg=%s, error=%s" % (ret['code'], ret['message'], errmsg), ret['code'])
        try:
            root = ET.fromstring(ret['body'])
        except ExpatError, e:
            raise MQEError("MQE failure: Result XML parse failure. code=%s, line=%s, offset=%s" % (e.code, e.lineno, e.offset))
        status = root.find('status')
        resultset = root.find('resultSet')
        if status.attrib['code'] == "COMPLETE":
            self.completed = True
        else:
            # Set last result
            objects = resultset.findall('object')
            # <resultSet> should have one or more <object> as the status is INCOMPLETE
            lastobj = objects[-1]
            self.last_result = HCPMQEOperationQuery.LastResultEntry(lastobj.attrib['urlName'], lastobj.attrib['changeTimeMilliseconds'], lastobj.attrib['version'])

        return resultset

class HCPMQEObjectQuery(object):
    def __init__(self):
        self.query = None
        self.queryexp = None
        self.count = -1
        self.offset = 0
        # Other parameters are not supported at the moment!
        self.hcpclient = None
        self.completed = False
        self.totalresults = 0

    def set_HCPClient(self, hcpclient):
        self.hcpclient = hcpclient

    def set_count(self, count):
        self.count = count

    def set_query(self, query):
        self.queryexp = query

    def set_offset(self, offset):
        self.offset = offset

    def build(self):
        # Top-level entry
        top = ET.Element("queryRequest")

        # object entry
        obj = ET.SubElement(top, "object")

        queryexp = ET.SubElement(obj, "query")
        queryexp.text = self.queryexp
        count = ET.SubElement(obj, "count")
        count.text = str(self.count)
        offset = ET.SubElement(obj, "offset")
        offset.text = str(self.offset)
        self.query = ET.tostring(top)
        return self.query

    def issue(self, build=True):
        if self.query is None or build is True:
            self.build()
        ret = self.hcpclient.exec_MQE(self.query)
        if ret['code'] != 200:
            if 'X-HCP-ErrorMessage' in ret['headers']:
                errmsg = ret['headers']['X-HCP-ErrorMessage']
            else:
                errmsg = "None"
            raise MQEError("MQE failure: HTTP return code=%s, msg=%s, error=%s" % (ret['code'], ret['message'], errmsg), ret['code'])
        try:
            root = ET.fromstring(ret['body'])
        except ExpatError, e:
            raise MQEError("MQE failure: Result XML parse failure. code=%s, line=%s, offset=%s" % (e.code, e.lineno, e.offset))
        status = root.find('status')
        resultset = root.find('resultSet')
        self.totalresults = int(status.attrib['totalResults'])
        if status.attrib['code'] == "COMPLETE":
            self.completed = True
        else:
            # Prepare for paged query
            self.offset += int(status.attrib['results'])

        return resultset

    def get_total_count(self):
        self.count = 0
        self.issue()
        return self.totalresults
