#!/usr/bin/python

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "generate_sshkeys"

import sys
import os
import getopt
import shutil

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions

key_comment = 'foscripts'
default_key_comment = key_comment + '-default'

def usage():
    print
    print "Usage:"
    print "    sudo maintexec %s --generate <configuration group name> [--override]" % PROGRAM
    print "    sudo maintexec %s --register <configuration group name>" % PROGRAM
    print "    sudo maintexec %s --remove <configuration group name>" % PROGRAM
    print

def get_registered_keys():
    """ Get a list of registered keys on this node.

    Args:
        None
    Returns:
        list: A list of comments of the registered keys on this node.
    """
    # Get the registered keys
    log_info("Getting a list of SSH keys registered on the node...")
    cmd = "sudo sshkeylist"
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = Utility.run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))

    if ret != 0:
        log_error("Failed to get a list of SSH keys on the node. ret=%s, out=%s, err=%s" % (ret, out, err))
        raise Exceptions.OperationError(cmd, ret)

    lines = out.split('\n')
    keys = [line for line in lines if len(line) != 0]
    comments = [key.strip().split()[2] for key in keys]

    return comments

def remove_ssh_key_if_exists(ssh_config, is_default=False):
    """ Remove the SSH key for the scripts if it exist.

    Get the currently registered SSH keys and check if the key for the
    scripts is already registered.
    If it's registered, remove it. If we are removing Unique keys, remove
    the key files as well. Otherwise copying a new key will fail due to
    permission error.

    This function should be called before calling generate_keys() and
    register_ssh_key(), otherwise both functions will fail due to the
    conflicts.

    Args:
        ssh_config (Config.SSHConfig): Configuration of SSH keys.
        is_default (bool): True if deleting a default key. False if
            deleting a Unique key.
    Returns:
        None
    """
    # Get the registered keys
    comments = get_registered_keys()

    # Remove the key if already registered
    if is_default:
        comment_to_be_deleted = default_key_comment
    elif ssh_config.confgroup is not None:
        comment_to_be_deleted = key_comment + '-' + ssh_config.confgroup.replace('_', '-')
    else:
        comment_to_be_deleted = key_comment
    if comment_to_be_deleted in comments:
        log_info("A key with the comment (%s) is already registered. Removing the key..." % comment_to_be_deleted)
        cmd = 'sudo sshkeydel -c "%s"' % comment_to_be_deleted
        log_debug("Command-Exec-Start: %s" % cmd)
        ret, out, err = Utility.run(cmd)
        log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))
        if ret != 0:
            log_error("Failed to remove the SSH key on the node. ret=%s, out=%s, err=%s" % (ret, out, err))
            raise Exceptions.OperationError(cmd, ret)

    # Remove the key file if deleting Unique keys to clean up things.
    # Do not delete default key files!
    if not is_default:
        log_info("Removing the existing key files of Unique SSH key ...")
        Utility.remove_file(ssh_config.get_public_key())
        Utility.remove_file(ssh_config.get_private_key())

def register_ssh_key(public_key_path, is_default=False, confgroup=None):
    """ Register an SSH public key on the node.

    This function registers an SSH public key specified by the argument
    on the node.
    remove_ssh_key_if_exists() should be called before calling this function
    to clean up SSH key registration status. Otherwise sshkeyadd will fail.

    Args:
        public_key_path (str): Path of the public key to register.
        is_default (bool): True if registering default SSH key. False if
            registering a Unique SSH key.
        confgroup (str): Configuration group name of the SSH keys to register.

    Returns:
        None: Nothing is returned.
    """
    if is_default:
        log_info("Registering a SSH public key as a default key. key=%s" % public_key_path)
    else:
        log_info("Registering a SSH public key. key=%s" % public_key_path)

    ### Register the public key

    # Copy the public key to /home/nasroot
    try:
        shutil.copy(public_key_path, "/home/nasroot")
    except EnvironmentError, e:
        log_error("Failed to copy SSH public key to /home/nasroot. error=%s" % str(e))
        raise

    # Register the public key
    key_name = os.path.basename(public_key_path)
    if is_default:
        comment = default_key_comment
    elif confgroup is not None:
        comment = key_comment + '-' + confgroup.replace('_', '-')
    else:
        comment = key_comment
    cmd = 'sudo sshkeyadd -c "%s" %s' % (comment, key_name)
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = Utility.run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))

    if ret != 0:
        log_error("Failed to register the SSH key on the node. ret=%s, out=%s, err=%s" % (ret, out, err))
        # Try to remove the copied key
        log_info("Deleting the key file on /home/nasroot ...")
        Utility.remove_file(os.path.join("/home/nasroot", key_name))
        raise Exceptions.OperationError(cmd, ret)

def generate_keys(work_dir, public_key_path, private_key_path, confgroup=None, override=False):
    """ Generate an SSH key pair on a specified directory.

    If confgroup is specified, confgroup name is added to key comment.
    If override is True, SSH keys are forcibly generated even if the
    keys already exist. If False, this function does not generate if
    the keys already exist.

    Args:
        work_dir (str): Working directory to store SSH keys.
        public_key_path (str): Path of a public key to generate.
        private_key_path (str): Path of a private key to generate.
        confgroup (str): Configuration group name for the SSH key to generate.
        override (bool): If this is True, override existing keys.

    Returns:
        None
    """
    log_info("Generating an SSH key pair.")

    tmp_key_name = "tmpkey"
    tmp_key_path = os.path.join(work_dir, tmp_key_name)

    # Check if the keys already exist
    if not override:
        if os.path.exists(private_key_path) and os.path.exists(public_key_path):
            # Already exists. Do nothing.
            log_info("SSH keys already exist.")
            return

    # Generate SSH key
    if confgroup is not None:
        comment = key_comment + '-' + confgroup.replace('_', '-')
    else:
        comment = key_comment
    cmd = 'ssh-keygen -t rsa -N "" -C "%s" -f %s' % (comment, tmp_key_path)
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = Utility.run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))
    if ret != 0:
        log_error("ssh-keygen failed with an error. ret=%s, out=%s, err=%s" % (ret, out, err))
        raise Exceptions.OperationError("cmd=%s, err=%s" % (cmd, err), ret)

    # Rename the private key
    try:
        os.rename(tmp_key_path, private_key_path)
    except EnvironmentError, e:
        log_error("Failed to rename SSH private key from %s to %s. error=%s" % (tmp_key_path, private_key_path, str(e)))
        raise

    # Set permission to the private key
    try:
        os.chmod(private_key_path, 0400)
    except EnvironmentError, e:
        log_error("Failed to set permission to SSH private key %s. error=%s" % (private_key_path, str(e)))
        raise

    # Rename the public key
    try:
        os.rename(tmp_key_path + '.pub', public_key_path)
    except EnvironmentError, e:
        log_error("Failed to rename SSH public key from %s to %s. error=%s" % (tmp_key_path + '.pub', private_key_path, str(e)))
        raise

    log_info("An SSH key pair was generated.")

def main():
    # Load config file
    config = Config.Config()

    # Parse options
    try:
        short_opts = ""
        long_opts = ["register=", "override", "remove=", "generate="]
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt.GetoptError, e:
        raise Exceptions.OptionError(e)

    generate = False
    override = False
    register = False
    remove = False
    confgroup = None
    for o, a in opts:
        if o == "--generate":
            generate = True
            confgroup = a
        if o == "--override":
            override = True
        if o == "--register":
            register = True
            confgroup = a
        if o == "--remove":
            remove = True
            confgroup = a

    if sum((generate, register, remove)) != 1:
        print "One of --generate, --register or --remove can be specified."
        raise Exceptions.OptionError("One of --generate, --register or --remove can be specified.")

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    ssh_config = Config.SSHConfig(config.sys_conf_dir, confgroup=confgroup)
    if register:
        # Register Unique public key
        print "Registering Unique key ..."
        register_ssh_key(ssh_config.get_public_key(), confgroup=confgroup)
    elif remove:
        # Remove Unique key
        print "Removing Unique key ..."
        remove_ssh_key_if_exists(ssh_config)
    elif generate:
        # Generate ssh keys
        print "Generating a Unique key pair ..."
        generate_keys(config.sys_conf_dir, ssh_config.get_public_key(), ssh_config.get_private_key(), confgroup=confgroup, override=override)
    print
    print "Completed."

if __name__ == '__main__':
    try:
        main()
    except Exceptions.OptionError, e:
        log_error("Option error: %s" % str(e))
        sys.exit(1)
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(10)
    except Exceptions.OperationError, e:
        log_exception("An operation error occurred: %s", e)
        sys.exit(90)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
