#!/usr/bin/python
from __future__ import with_statement

BASEDIR = "/home/nasroot/foscripts"
PROGRAM = "setup_scripts"

import sys
import os
import errno
import base64
import hashlib

sys.path.append(os.path.join(BASEDIR, "bin/folib"))
import Utility
import HDIConfig
import HDIClient
import HDIOperation
from CommonLogger import log_debug, log_info, log_warning, log_error, log_exception, setup_rotation_logger

sys.path.append(os.path.join(BASEDIR, "bin"))
import Config
import Exceptions
import generate_sshkeys
import getinfo_common

def test_api_password(config, passMD5):
    """ Test MAPI request is authenticated by the specified password.

    Issue MAPI GET request to /mapi/APIAccount as a testing, with the specified
    password (MD5 hashed).
    Returns True if it succeeded (200 is returned). Returns False if authentication
    failed (401 is returned).
    Other error will end up with an exception.

    Args:
        config (Config.Config): Configuration of the scripts.
        passMD5 (str): MD5 hashed password to be tested.

    Returns:
        bool: True if succeeded. False if failed. Otherwise an exception will be
            thrown.
    """
    # Setup MAPI connection
    cred = HDIClient.Credential(None, config.rest_user, None, passMD5)
    conn = HDIOperation.Connection("localhost", cred, config.retry_count, config.retry_interval, config.timeout)

    log_info("Testing API password ...")
    method = HDIClient.HTTPMethod.GET
    path = "/mapi/APIAccount"
    op = HDIOperation.RESTOperation(method, path)

    # Issue MAPI
    try:
        ret, out, err = op.execute(conn)
    except HDIOperation.ConnectionError, e:
        log_error("Failed to communicate with %s. error=%s" % (conn.host, str(e)))
        raise Exceptions.ConnectionError(e.reason, e.code)

    # Check result
    if ret == 200:
        log_info("API execution succeeded.")
        return True
    elif ret == 401:
        log_info("API execution failed with authentication error.")
        return False
    else:
        log_error("Unexpected HTTP status code. host=%s, path=%s, method=%s, ret=%s, out=%s, err=%s" % (conn.host, path, HDIClient.HTTPMethod.tostr(method), ret, out, err))
        raise Exceptions.OperationError("MAPI execution failed. host=%s, error=%s" % (conn.host, err), ret)

def setup_api_password(config):
    log_info("Setting up API password ...")

    # Try the current configured password
    print "Validating API password ... ",
    result = test_api_password(config, config.rest_passMD5)
    if result is True:
        print "OK."
        log_info("The currently configured password is validated.")
        # Finish
        return
    else:
        print "NG."
        log_info("The currently configured password is not correct.")

    # Ask a correct password if test did not pass
    # Get model information of this node
    print
    print "Getting model information of this node ..."
    model = getinfo_common.get_local_model(config)
    if model == HDIConfig.Model.CLUSTER:
        prompt = "Please input API password of this node: "
    elif model == HDIConfig.Model.SINGLE:
        prompt = "Please input GUI login password of this node: "

    # Ask a password and try validation again
    retry = 2
    tried = 0
    while tried <= retry:
        tried += 1

        # Ask a password
        password = Utility.my_input_secret(prompt)
        if password is None:
            # EOF is entered
            raise Exceptions.NoExecution()
        if len(password) == 0:
            # Only 'enter' was input - try again
            continue

        # Make MD5 hashed password
        m = hashlib.md5()
        m.update(password)
        passMD5 = m.hexdigest()

        # Test with the new password
        print "Validating API password ... ",
        result = test_api_password(config, passMD5)
        if result is True:
            print "OK."
            log_info("The input password is validated.")
            break
        else:
            print "NG."
            log_info("The input password is not correct.")
    else:
        print "A wrong password was input too many times."
        raise Exceptions.NoExecution()

    # Update API credential configuration file.
    print "Updating configuration file ..."
    user64 = base64.b64encode(config.rest_user)
    try:
        with open(config.hdi_auth_path, 'w') as f:
            f.write("%s:%s" % (user64, passMD5))
    except EnvironmentError, e:
        log_error("Failed to update API configuration file. error=%s" % str(e))
        raise

def setup_system_UUID():
    """ Setup system UUID if not configured yet.

    System UUID is recorded /enas/hsmarc/conf/system_uuid.conf.
    First check the existence of system_uuid.conf. If it doesn't exist,
    execute 'sudo syslusave -d trans' to generate system UUID.

    Args:
        None
    Returns:
        None
    """
    system_uuid_path = "/enas/hsmarc/conf/system_uuid.conf"

    # Check the existence of system_uuid.conf
    if not os.path.exists(system_uuid_path):
        log_info("System UUID is not configured yet.")
        print "System UUID is not configured yet."

        # Ask the admin if they want to execute syslusave now
        answer = Utility.yes_or_no_full("Do you want to execute system backup (syslusave) now to configure system UUID?")
        if answer is True:
            # Execute syslusave
            cmd = "sudo syslusave -d trans"
            log_debug("Command-Exec-Start: %s" % cmd)
            ret, out, err = Utility.run(cmd)
            log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))

            if ret != 0:
                log_error("Failed to execute system backup. ret=%s, out=%s, err=%s" % (ret, out, err))
                print "Failed to execute system backup."
                print "Please execute 'syslusave -d trans' manually later, or just wait for the next daily system backup."
                return

            # Check again
            if not os.path.exists(system_uuid_path):
                log_error("%s does not exist even after executing syslusave." % system_uuid_path)
                print "It seems something failed while executing system backup."
                print "Please execute 'syslusave -d trans' manually later, or just wait for the next daily system backup."
        else:
            # The admin doesn't want to execute syslusave now
            # Show some guidance
            log_info("Execution of syslusave was rejected by the user.")
            print "OK, please execute 'syslusave -d trans' manually later, or just wait for the next daily system backup."
    else:
        log_info("System UUID is already configured.")
        print "System UUID is already configured."

def setup_log_collection():
    log_info("Executing log collection setup script ...")
    print "Executing log collection setup script ..."

    log_setup_prog = "int/setup_log_collection"
    cmd = "sudo maintexec %s" % os.path.join(BASEDIR, log_setup_prog)
    log_debug("Command-Exec-Start: %s" % cmd)
    ret, out, err = Utility.run(cmd)
    log_debug("Command-Exec-End: %s, ret=%s" % (cmd, ret))
    if ret != 0:
        log_error("Failed to setup log collection script. cmd=%s, ret=%s, out=%s, err=%s" % (cmd, ret, out, err))
        raise Exceptions.OperationError("Log collection setup failed.", ret)

def main():
    ### Preparation

    # Enable Ctrl-C interruption
    Utility.enable_interrupt()

    # Load configuration file
    config = Config.Config()

    # Setup logging
    log_file_path = os.path.join(config.log_dir, config.common_log_file)
    setup_rotation_logger(__name__, config.log_level, log_file_path, config.log_size, config.log_gen)

    log_info("----------------------- %s --------------------------" % PROGRAM)

    # Setup default SSH keys
    ssh_config = Config.SSHConfig(config.sys_conf_dir)
    print
    print "----- Default SSH key setup -----"
    print "Checking status of SSH key registration and cleaning up ..."
    generate_sshkeys.remove_ssh_key_if_exists(None, is_default=True)
    print "Registering default SSH key ..."
    generate_sshkeys.register_ssh_key(ssh_config.get_public_key(is_default=True), is_default=True)

    # Setup API password
    print
    print "----- API password setup -----"
    setup_api_password(config)

    # Setup system UUID
    print
    print "----- System UUID setup -----"
    setup_system_UUID()

    # Setup log collection script
    print
    print "----- Log collection setup -----"
    setup_log_collection()

    # End
    print
    print "Setup completed."

if __name__ == '__main__':
    try:
        main()
    except Config.BaseError, e:
        print "Unable to load script configuration: %s" % str(e)
        sys.exit(1)
    except Exceptions.NoExecution:
        print "Process canceled."
        sys.exit(1)
    except Exceptions.OperationError, e:
        log_error("An operation error occurred: %s" % str(e))
        sys.exit(99)
    except EnvironmentError, e:
        if e.errno == errno.ENOSPC:
            log_error("There is no space on the device. error=%s" % str(e))
            sys.exit(70)
        else:
            log_exception("An internal error occurred: %s", e)
            sys.exit(99)
    except Exception, e:
        log_exception("An internal error occurred: %s", e)
        sys.exit(99)

    sys.exit(0)
